"""Main module."""

from itertools import chain

from .namefinder import NumberNameFinder


class UnoccupiedNameNotFound(Exception):
    """Raise when unoccupied name not found."""


def unoccupied(
        basename, occupied, name_finder=NumberNameFinder(), nobase=False):
    """Get unoccupied name by base name, occupied names & name gen func.

    Args:
        basename: (str)
            base name (the wanted name)
        occupied: (container of str)
            the names already be occupied.
        name_finder: (callable)
            unoccupied name generator.

            Signature:
                name_finder(basename, occupied) -> unoccupied_name

                `name_finder()` must return str type or None.

                If return None, mean not found any unoccupied name.
                    ==> raise UnoccupiedNameNotFound exception

                Note: occupied already be convert to `frozenset` data type.

        nobase: (bool)
            not allow use `basename` directly. even basename not in occupied.

    Returns:
        basename, if basename already be occupied, return a unoccupied_name

    """
    if isinstance(occupied, str):
        raise TypeError('`occupied` can not a str type object.')

    if nobase:
        norm_occupied = frozenset(chain([basename], occupied))
    else:
        norm_occupied = frozenset(occupied)

    if basename not in norm_occupied:
        return basename
    else:
        new_name = name_finder(basename, norm_occupied)

    if isinstance(new_name, str):
        if new_name in norm_occupied:
            raise UnoccupiedNameNotFound(
                '`name_finder` should not return a name "{}"'
                ' already in occupied.'
                .format(new_name)
            )

        return new_name

    elif new_name is None:
        raise UnoccupiedNameNotFound('can not found a unoccupied name.')

    else:
        raise TypeError('name_finder ({}) not return a `str` but {}.'
                        .format(name_finder.__name__,
                                type(new_name).__name__))



"""Test main module."""

import pytest

from src.unoccupied.main import unoccupied
from src.unoccupied.main import UnoccupiedNameNotFound


class TestUnoccupied:
    @pytest.mark.parametrize('basename, occupied, expected', [
        ('foo', [], 'foo'),
        ('foo', ['bar'], 'foo'),
        ('foo', ['foo'], 'foo-1'),
    ])
    def test_basic(self, basename, occupied, expected):
        assert unoccupied(basename, occupied) == expected

    def test_occupied_is_str(self):
        with pytest.raises(TypeError):
            unoccupied('foo', 'bar')  # str as `occupied` are not allow

    def test_nobase(self):
        assert unoccupied('foo', [], nobase=True) == 'foo-1'

    def test_call_name_finder_only_needed(self):
        called = False

        def pseudo_name_finder(basename, occupied):
            nonlocal called
            called = True

            return 'whatever'

        unoccupied('foo', [], pseudo_name_finder)
        assert called is False

        unoccupied('foo', ['foo'], pseudo_name_finder)
        assert called is True

    def test_occupied_become_frozenset(self):
        def pseudo_name_finder(basename, occupied):
            assert isinstance(occupied, frozenset)

            return 'whatever'

        unoccupied('foo', ['foo'], pseudo_name_finder)

    def test_name_finder_return_none(self):
        with pytest.raises(UnoccupiedNameNotFound):
            def pseudo_name_finder(basename, occupied):
                return None

            unoccupied('foo', ['foo'], pseudo_name_finder)

    def test_name_finder_return_basename(self):
        with pytest.raises(UnoccupiedNameNotFound):
            def pseudo_name_finder(basename, occupied):
                return basename

            unoccupied('foo', ['foo'], pseudo_name_finder)

    def test_name_finder_return_elem_in_occupied(self):
        with pytest.raises(UnoccupiedNameNotFound):
            def pseudo_name_finder(basename, occupied):
                return 'bar'

            unoccupied('foo', ['foo', 'bar'], pseudo_name_finder)

    def test_name_finder_return_elem_not_str_or_none(self):
        with pytest.raises(TypeError):
            def pseudo_name_finder(basename, occupied):
                return ['return a list']

            unoccupied('foo', ['foo', 'bar'], pseudo_name_finder)



"""Name finder - find and return a unoccupied name."""

import os

from itertools import count


class NumberNameFinder:
    """Use `basename` and increasing `num` to find a new name."""

    def __init__(self, template='{basename}-{num}', start=1):
        """Configure NumberNameFinder.

        Args:
            template: (str)
                a python str.format template.
                this template can include 2 keyword params:
                    {basename}: wanted basename
                    {num}: increasing number

            start: (int)
                start testing value for {num} in template.
        """
        self.template = template
        self.start = start

    def __call__(self, basename, occupied):
        """Get the first unoccupied name (not include `basename` itself)."""
        checked = set()

        for num in count(start=self.start):
            testing_name = self.template.format(basename=basename, num=num)

            if testing_name in checked:
                return None
            else:
                checked.add(testing_name)

            if testing_name not in occupied:
                return testing_name


class FileNameFinder:
    """Auto process filename extension then find a new name."""

    def __init__(self, template='{base}-{num}', start=1):
        """Configure FileNameFinder.

        Args:
            template: (str)
                a python str.format template.
                this template can include 2 keyword params:
                    {base}: filename base part from wanted filename
                    {num}: increasing number

                Result will be auto append the original filename extension.

            start: (int)
                start testing value for {num} in template.
        """
        self.template = template
        self.start = start

    def __call__(self, basename, occupied):
        """Get the first unoccupied name (not include `basename` itself)."""
        checked = set()
        base, ext = os.path.splitext(basename)

        for num in count(start=self.start):
            testing_name = self.template.format(base=base, num=num) + ext

            if testing_name in checked:
                return None
            else:
                checked.add(testing_name)

            if testing_name not in occupied:
                return testing_name



"""Cmdlr infomation files."""

VERSION = (0, 1, 1)

VERSION_STR = '.'.join(map(lambda x: str(x), VERSION))
DESCRIPTION = 'Generate unoccupied name.'
LICENSE = 'MIT'
AUTHOR = 'visig'
AUTHOR_EMAIL = 'visig@protonmail.ch'
PROJECT_URL = 'https://github.com/civalin/unoccupied'
PROJECT_NAME = 'unoccupied'



#!/usr/bin/env python3

"""Install script."""

from setuptools import setup, find_packages
import src.unoccupied.info as info

with open('README.md', mode='r', encoding='utf8') as f:
    long_desc = f.read()

setup(
    name=info.PROJECT_NAME,
    version=info.VERSION_STR,
    author=info.AUTHOR,
    author_email=info.AUTHOR_EMAIL,
    license=info.LICENSE,
    url=info.PROJECT_URL,
    description=info.DESCRIPTION,
    long_description=long_desc,
    long_description_content_type='text/markdown',

    classifiers=[
        'Programming Language :: Python :: 3',
    ],
    install_requires=[],
    setup_requires=['pytest-runner'],
    tests_require=['pytest >=3.6, <4'],
    package_dir={'': 'src'},
    packages=find_packages('src'),
    include_package_data=True,
)



"""Test namefinder module."""

import pytest

from src.unoccupied.namefinder import NumberNameFinder
from src.unoccupied.namefinder import FileNameFinder


class TestNumberNameFinder:
    @pytest.mark.parametrize('template, start, expected', [
        ('{basename}', 0, None),
        ('{basename}-{num}', 0, 'foo-0'),
        ('{basename}-{num}', 1, 'foo-1'),
        ('{basename}{num}', 2, 'foo2'),
        ('{basename}.{num}', 2, 'foo.2'),
        ('{num}-{basename}', 2, '2-foo'),
        ('{basename}-{num:04}', 2, 'foo-0002'),
    ])
    def test_define(self, template, start, expected):
        number_name_finder = NumberNameFinder(template=template, start=start)

        basename = 'foo'
        occupied = ['foo']

        assert number_name_finder(basename, occupied) == expected

    @pytest.mark.parametrize('basename, occupied, expected', [
        ('foo', ['foo'], 'foo-1'),
        ('foo', ['foo', 'foo-1'], 'foo-2'),
        ('foo', ['foo', 'foo-2'], 'foo-1'),
    ])
    def test_usage_with_default_define(self, basename, occupied, expected):
        number_name_finder = NumberNameFinder()

        assert number_name_finder(basename, occupied) == expected


class TestFileNameFinder:
    @pytest.mark.parametrize('template, start, expected', [
        ('{base}', 0, None),
        ('{base}-{num}', 0, 'foo-0.txt'),
        ('{base}-{num}', 1, 'foo-1.txt'),
        ('{base}{num}', 2, 'foo2.txt'),
        ('{base}.{num}', 2, 'foo.2.txt'),
        ('{num}-{base}', 2, '2-foo.txt'),
        ('{base}-{num:04}', 2, 'foo-0002.txt'),
    ])
    def test_define(self, template, start, expected):
        file_name_finder = FileNameFinder(template=template, start=start)

        basename = 'foo.txt'
        occupied = ['foo.txt']

        assert file_name_finder(basename, occupied) == expected

    @pytest.mark.parametrize('basename, occupied, expected', [
        ('foo.txt', ['foo.txt'], 'foo-1.txt'),
        ('foo.txt', ['foo.txt', 'foo-1.txt'], 'foo-2.txt'),
        ('foo.txt', ['foo.txt', 'foo-2.txt'], 'foo-1.txt'),
    ])
    def test_usage_with_default_define(self, basename, occupied, expected):
        file_name_finder = FileNameFinder()

        assert file_name_finder(basename, occupied) == expected



"""Unoccupied - Find unoccupied name."""

from .main import unoccupied
from .main import UnoccupiedNameNotFound
from .namefinder import NumberNameFinder
from .namefinder import FileNameFinder
from .info import VERSION as __version__  # NOQA

__all__ = ['unoccupied',
           'UnoccupiedNameNotFound',
           'NumberNameFinder',
           'FileNameFinder']
